/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "group.h"

#include <assert.h>

#include <iostream>
#include <algorithm>

namespace Papyrus
{

  Group::Group()
  {
    set_selectable( false );
  }

  Group::pointer Group::create()
  {
    return Group::pointer( new Group() );
  }

  Group::~Group()
  {
    Children::iterator childiter;
    Connections::iterator conniter;

    for ( conniter = m_redraw_connections.begin(); conniter != m_redraw_connections.end(); conniter++ )
      conniter->second.disconnect();

    for ( childiter = m_children.begin(); childiter != m_children.end(); childiter++ )
      ( *childiter ) ->m_parent = NULL;
  }

  bool Group::add( Drawable::pointer item, bool drawing_only )
  {
    Connections::iterator i;

    // We must have a valid pointer
    if ( ! item ) return false;

    // we can't add an item to itself, otherwise we'll end up with recursive problems
    if ( item.get() == this )
      return false;

    if ( drawing_only ) {
      
      // if we're already the drawing parent, we're done
      if ( item->m_drawing_parent == this )
        return true;
      
      // if it already has a drawing parent, let's disconnect it from need_redraw messages
      if ( item->m_drawing_parent ) {
        i = m_redraw_connections.find( item );
        if ( i != m_redraw_connections.end() ) {
          i->second.disconnect();
          m_redraw_connections.erase( i );
        }
      }
      // this group becomes the new drawing parent
      item->m_drawing_parent = this;

      // we need the redraw signal connected
      // we'll connect the changed signal later
      m_redraw_connections[ item ] =
          item->signal_redraw().connect( sigc::bind( sigc::mem_fun( *this, &Group::on_child_need_redraw ), item ) );

    } else { // not drawing only, but a normal add

      // if we're already the parent, we're done
      if ( item->m_parent == this )
        return true;

      // if it already has a parent, we need to remove it from the old parent
      if ( item->m_parent != NULL )
        item->m_parent->remove( item );

      // now, let's make this group the parent
      item->m_parent = this;

      // if it doesn't already have a drawing parent, we'll make this the drawing parent
      if ( item->m_drawing_parent == NULL ) {
        item->m_drawing_parent = this;

        // we need the redraw signal connected
        // we'll connect the changed signal later
        m_redraw_connections[ item ] =
            item->signal_redraw().connect( sigc::bind( sigc::mem_fun( *this, &Group::on_child_need_redraw ), item ) );
      }
    }

    // Add this object to the list of children
    m_children.push_front( item );
    
    // Connect to the changed signal
    m_changed_connections[ item ] =
        item->signal_changed().connect( sigc::bind( sigc::mem_fun( *this, &Group::on_child_changed ), item ) );

    // Include the child's extents in this group's extents
    m_extents.include( item->global_extents() );

    // Perform any work needed to update the extents
    this->update_extents();

    // Notify anyone connected to the child_added signal
    m_signal_child_added.emit( item );

    // We'll need a redraw of course
    this->need_redraw();

    return true;
  }

  bool Group::remove( Drawable::pointer item )
  {
    bool removed = false;
    Connections::iterator i;

    // We must have a valid pointer
    if ( ! item )
      return false;

    // Whether the child is in the list or not, let's just try to remove it
    m_children.remove( item );

    // Let's find out if we have a redraw signal for it
    i = m_redraw_connections.find( item );
    if ( i != m_redraw_connections.end() ) {
      // if we do, let's disconnect and remove the connection
      i->second.disconnect();
      m_redraw_connections.erase( i );
      removed = true;
    }

    // Let's find out if we have a changed signal connection
    i = m_changed_connections.find( item );
    if ( i != m_changed_connections.end() ) {
      // if we do, let's disconnect and remove the connection
      i->second.disconnect();
      m_changed_connections.erase( i );
      removed = true;
    }

    // find out if we were the drawing parent only
    // if we were, let's make the real parent the drawing parent
    if ( item->m_drawing_parent == this && item->m_parent != this ) {
      item->m_drawing_parent = item->m_parent;
      if ( item->m_drawing_parent ) {
        item->m_drawing_parent->m_redraw_connections[ item ] =
            item->signal_redraw().connect( sigc::bind( sigc::mem_fun( *(item->m_drawing_parent), &Group::on_child_need_redraw ), item )  );
        item->m_drawing_parent->need_redraw();
      }
    }
    // if we weren't the drawing parent, let's see if we're the regular parent
    else if ( item->m_parent == this ) 
      item->m_parent = NULL;
    
    if ( removed ) {
      this->update_extents();
      m_signal_child_removed.emit( item );
      this->need_redraw();
    }
    
    return removed;
  }

  bool Group::remove( Drawable* object )
  {
    Children::iterator i;

    // We must have a valid pointer
    if ( object == NULL )
      return false;

    // Let's look for the pointer and then call the remove method with the smart pointer
    for ( i = m_children.begin(); i != m_children.end(); i++ ) {
      if ( i->get() == object )
        return this->remove( *i );
    }
    return false;
  }

  bool Group::clear()
  {
    while ( m_children.size() > 0 )
      remove( m_children.front() );
    update_extents();
    this->need_redraw();
    return true;
  }

  void Group::draw( Cairo::RefPtr<Cairo::Context> cairo )
  {
    Children::reverse_iterator i;
    for ( i = m_children.rbegin(); i != m_children.rend(); i++ )
      if ( (*i)->m_drawing_parent == this )
        ( *i ) ->render( cairo );
  }

  void Group::on_child_need_redraw( double x, double y, double w, double h, Drawable::pointer child )
  {
    Region region( x, y, w, h );
    recalculate_matrix();
    region.transform( m_matrix );
    region.xywh( x, y, w, h );
    m_signal_redraw.emit( x, y, w, h );
  }

  bool Group::raise( Drawable::pointer item )
  {
    Children::iterator i, j;
    j = std::find( m_children.begin(), m_children.end(), item );
    if ( j != m_children.end() && j != m_children.begin() ) {
      i = j;
      i--;
      m_children.erase( j );
      m_children.insert( i, item );
      item->need_redraw();
      return true;
    }
    return false;
  }

  bool Group::raise_to_top( Drawable::pointer item )
  {
    Children::iterator i;
    i = std::find( m_children.begin(), m_children.end(), item );
    if ( i != m_children.end() && i != m_children.begin() ) {
      m_children.erase( i );
      m_children.push_front( item );
      item->need_redraw();
      return true;
    }
    return false;
  }

  bool Group::lower( Drawable::pointer item )
  {
    Children::iterator i, j;
    i = std::find( m_children.begin(), m_children.end(), item );
    if ( i != m_children.end() && ( item != m_children.back() ) ) {
      j = i;
      j++;
      j++;
      m_children.erase( i );
      m_children.insert( j, item );
      item->need_redraw();
      return true;
    }
    return false;
  }

  bool Group::lower_to_bottom( Drawable::pointer item )
  {
    Children::iterator i;
    i = std::find( m_children.begin(), m_children.end(), item );
    if ( i != m_children.end() && ( item != m_children.back() ) ) {
      m_children.erase( i );
      m_children.push_back( item );
      item->need_redraw();
      return true;
    }
    return false;
  }

  bool Group::inside( double x, double y )
  {
    if ( is_selectable() )
      return Drawable::inside( x, y );
    for ( Children::iterator i = m_children.begin(); i != m_children.end(); i++ )
      if ( ( *i ) ->inside( x, y ) )
        return true;
    return false;
  }

  sigc::signal<void, Drawable::pointer>& Group::signal_child_added()
  {
    return m_signal_child_added;
  }

  sigc::signal<void, Drawable::pointer>& Group::signal_child_removed()
  {
    return m_signal_child_removed;
  }

  bool Group::is_group()
  {
    return true;
  }

  std::vector<Drawable::pointer> Group::select( double x, double y, unsigned depth )
  {
    std::vector<Drawable::pointer> result, intermediate;
    Group::pointer g;

    if ( depth == 0 )
      return result;

    //     // it's not worth checking the children individually
    //     // if the select point isn't even in the group extents
    //   if (get_global_extents().inside(x, y)) {

    // transform the points according to the group
    // affine matrix and check each of the children
    cairo_matrix_t inverse = matrix();
    cairo_matrix_invert( &inverse );
    cairo_matrix_transform_point( &inverse, &x, &y );
    for ( Children::iterator i = m_children.begin(); i != m_children.end(); i++ ) {
      if ( ( *i ) ->inside( x, y ) ) {
        if ( ( *i ) ->is_group() ) {
          g = papyrus_smart_pointer_cast<Group>( ( *i ) );
          if ( g->is_selectable() ) {
            result.push_back( g );
            depth--;
          }
          intermediate = g->select( x, y, depth );
          result.insert( result.end(), intermediate.begin(), intermediate.end() );
        } else {
          result.push_back( *i );
          depth--;
        }
      }
      if ( depth == 0 )
        break;
    }
    //     }
    return result;
  }

  void Group::rebuild_extents()
  {
    m_extents.set_xywh( 0.0, 0.0, 0.0, 0.0 );
    if ( m_children.size() > 0 ) {
      m_extents = m_children.front() ->global_extents();
      for ( Children::iterator i = m_children.begin(); i != m_children.end(); i++ )
        m_extents.include( ( *i ) ->global_extents() );
    }
  }

  void Group::on_child_changed( Drawable::pointer )
  {
    //   if (changed & RENDER_BBOX)
    update_extents();
  }

  Group::Children& Group::get_children( )
  {
    return m_children;
  }


  void Group::update_extents( )
  {
    rebuild_extents();
    this->need_redraw();
  }

}

